#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['qpsolvers']

package_data = \
{'': ['*'], 'qpsolvers': ['.ropeproject/*']}

install_requires = \
['quadprog >= 0.1.8']

extras_require = \
{'doc': ['sphinx']}

setup(name='qpsolvers',
      version='1.6.1',
      description='Quadratic programming solvers in Python with a unified API',
      author='Stéphane Caron',
      author_email='stephane.caron@normalesup.org',
      url='https://github.com/stephane-caron/qpsolvers',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
